# coding=utf-8
"""
python t0 配平 mini qmt版
单账号版
"""

# mini python base library
from xtquant.xttrader import XtQuantTrader, XtQuantTraderCallback
from xtquant.xttype import StockAccount
from xtquant.xtdata import get_full_tick
from xtquant import xtconstant
import time

# 队列里的类型：[主推类型, 股票代码， 数量， 方向]

# stragegy library
import sys, re, traceback

if sys.version_info.major != 3:
    sys.exit('版本不受支持，请使用python3.x！')

import os

TRADE = 'trade'
ORDER = 'order'
CANCEL_ERROR = 'cancel_order'
ORDER_ERROR = 'order_error'

PART_CANCEL = 53
CANCELED = 54

import logging

appLogger = logging.getLogger(__name__ + 'python配平')
# 可以设置日志级别
appLogger.setLevel(logging.DEBUG)

# stderr流，控制日志输出到标准显示器
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)

# file流，控制日志输出到指定路径
handler = logging.FileHandler(r'..\userdata_mini\log\python balance.log')
handler.setLevel(logging.DEBUG)

fmt = logging.Formatter(
    "[%(asctime)s] [%(name)s] [%(module)s] [%(lineno)d] [%(levelname)s] %(message)s")

fmt_stream = logging.Formatter(
    "[%(module)s] [%(levelname)s] %(message)s")

console.setFormatter(fmt_stream)
handler.setFormatter(fmt)

appLogger.addHandler(console)


appLogger.addHandler(handler)


class StrategyMniQmtT0(XtQuantTraderCallback):

    def __init__(self, over_order_price=0.02, trading_time=False, loop=10):
        """

        :param over_order_price:超价设置
        :param loop:轮询周期

        """
        ud_path, ud_session, account, self._strategy_name, stocks, volumes, directions = parse_local_command(0)
        # 库基础配置
        self.acc = account

        # qmt 库初始化
        self.t = XtQuantTrader(ud_path, ud_session)
        self.acc = StockAccount(account)
        self.t.register_callback(self)
        self.t.start()
        r1 = self.t.connect()
        self.t.subscribe(self.acc)

        self.tzbz = time.strftime('%Y%m%d %H%M%S')  # 配平默认使用的投资备注
        if r1 != 0:
            raise ValueError('连接client 失败，错误状态:%d' % r1)

        # 参数
        self.over_order_price = over_order_price
        self.loop_time = loop
        # self.stocks = stocks                        # 股票
        # self.volumes = volumes                      # 量
        # self.volumes = volumes                      # 方向

        # 包装目标任务
        self.total_dict_buy = {}
        self.total_dict_sell = {}
        self.total_trade_list = list(set(stocks))

        if len(self.total_trade_list) != len(stocks):
            appLogger.critical('传入的配平股票有重复，请重试！')
            raise ValueError('传入的配平股票有重复，请重试！')

        for i, s in enumerate(stocks):
            if directions[i] == 1:  # 需要卖 配平
                self.total_dict_sell[s] = volumes[i]
            elif directions[i] == 0:  # 需要买 配平
                self.total_dict_buy[s] = volumes[i]
            else:
                appLogger.error('{}未知的买卖类型{}，跳过配平该股票'.format(s, directions[i]))
        # 记录委托状态的字典
        self.ordered_status = {}  # 根据股票唯一配平原则（即配平篮子中股票是唯一的，不存在一个股票需要2种配平的情况），key为股票，value为状态
        self.wrong_status = {}  # 失败状态，
        self.ready_to_buy = self.total_dict_buy.copy()  # 可买股票池
        self.ready_to_sell = self.total_dict_sell.copy()  # 可卖股票池

        self.ordered = {}  # 记录下单股票和数量，以存储下单报错时的下单参数
        self.cancelable_list = [48, 49, 50, 51, 52, 55]

        # 变量
        self.__first_call = False
        self.trading_time = trading_time  # 非交易时间是否可继续配平
        # 程序内计时器
        self.__timetag = time.time
        self.__trade_time = self.__timetag()

    @property
    def __has_ordered(self):
        if not self.__first_call:
            self.__first_call = True
            return False
        else:
            return self.__first_call

    def __call_sleep_before_next_loop(self):
        self.__trade_time += self.loop_time

    def init(self):
        appLogger.info('配平开始'+'=='*30)
        self.__handlebar()

    def __handlebar(self):
        while 1:
            cloeker = self.__timetag() >= self.__trade_time  # 定期检查是否有未完成的任务，如果没有则程序退出
            # appLogger.info('looping handlebar')
            if time.strftime('%H%M%S') > '153002':
                if self.trading_time:
                    appLogger.error('交易时间已过！不再配平')
                    if self.total_dict_buy:
                        appLogger.error('买方向剩余配平任务：{}'.format(self.total_dict_buy))
                    if self.total_dict_sell:
                        appLogger.error('卖方向剩余配平任务：{}'.format(self.total_dict_sell))
                    break
                else:
                    pass
            if self.__has_ordered:
                # r =
                self.update(cloeker)
            if not cloeker:
                continue
            if self.check_order_status():
                # print(self.total_dict_sell, self.total_dict_buy)
                # 可能会做其他工作，如结束任务，退出循环
                appLogger.error('配平任务完成')
                break

            appLogger.debug('剩余任务(卖):{}'.format(self.total_dict_sell))
            appLogger.debug('剩余任务(买):{}'.format(self.total_dict_buy))
            # print('准备卖：', self.ready_to_sell)
            # print('准备买：', self.ready_to_buy)
            # 下单
            if self.ready_to_buy:
                self.do_trade(self.ready_to_buy, xtconstant.STOCK_BUY)
            if self.ready_to_sell:
                self.do_trade(self.ready_to_sell, xtconstant.STOCK_SELL)
            self.__call_sleep_before_next_loop()
        appLogger.info('配平完成，任务结束！')

    def update(self, clocker):
        __ordered_status_list = list(self.ordered_status.keys())
        # dealed_stock = []
        for stock in __ordered_status_list:
            trade_info = self.ordered_status[stock]
            __type = trade_info[0]
            __obj = trade_info[1]
            stock = stock
            # if stock not in dealed_stock:
            #     dealed_stock.append(stock)
            # else:
            #     continue # 不需要旧的委托状态
            if __type == TRADE:  # 对于成交股票，减去相应任务数量
                del self.ordered_status[stock]
                # print('实际已成：',stock,__obj.traded_volume)
                if __obj.order_type == xtconstant.STOCK_SELL:
                    self.total_dict_sell[stock] -= __obj.traded_volume
                    if self.total_dict_sell[stock] == 0:  # 单品种任务完成 则从总任务中剔除
                        del self.total_dict_sell[stock]

                else:
                    self.total_dict_buy[stock] -= __obj.traded_volume
                    if self.total_dict_buy[stock] == 0:
                        del self.total_dict_buy[stock]

            if __type == ORDER and clocker:  # 对于仍未成交的股票：撤单
                # dealed_stock.append(__obj.stock_code)
                if __obj.order_status in self.cancelable_list:  # 如果可撤
                    if __obj.order_volume - __obj.traded_volume < 100:  # 未成交时０股时不撤单
                        # print('未成交数量剩0股，忽略',__obj.stock_code,__obj.order_volume - __obj.traded_volume)
                        del self.ordered_status[stock]
                        continue
                    else:
                        _cancel_symbol = self.do_cancel(__obj.order_id)  # 撤单
                        # print('未成交撤单',stock,__obj.order_volume - __obj.traded_volume)
                        del self.ordered_status[stock]
                        if _cancel_symbol < 0:  # 报单失败，忽略该委托
                            # appLogger.error('%s 撤单失败，忽略后续操作' % self.ordered[__obj.order_id][0])
                            # 对于xtquant层的撤单失败：
                            self.wrong_status[__obj.stock_code] = __obj
                else:
                    del self.ordered_status[stock]

        if not clocker:  # 以下时对错误状态的处理
            return
        error_list = list(self.wrong_status.keys())
        for stock in error_list:
            trade_info = self.wrong_status[stock]
            __type = trade_info[0]
            __obj = trade_info[1]
            # if __type == ORDER_ERROR : # 对于下单失败：
            #     if __obj.order_type==xtconstant.STOCK_BUY:
            #         self.ready_to_buy[stock] = __obj.order_volumer
            #         del self.wrong_status[stock]
            if __type == CANCEL_ERROR:  # 对于撤单失败，重新撤单
                appLogger.debug()
                _cancel_symbol = self.do_cancel(__obj.order_id)  # 撤单
                # print('撤单错误重新撤单', stock, __obj.order_volume - __obj.traded_volume)
                del self.wrong_status[stock]
                if _cancel_symbol < 0:  # 报单失败，忽略该委托
                    # appLogger.error('%s 撤单失败，忽略后续操作' % self.ordered[__obj.order_id][0])
                    # 对于xtquant层的撤单失败：
                    self.wrong_status[__obj.stock_code] = __obj

    def check_order_status(self):
        if (not self.total_dict_sell) and (not self.total_dict_buy):
            return True
        else:
            return False

    def do_cancel(self, order_id):
        return self.t.cancel_order_stock(self.acc, order_id)

    def do_trade(self, basket, types):
        quoter = list(basket.keys())
        market_data = get_full_tick(quoter)
        if types == xtconstant.STOCK_SELL:  # 卖
            for s in quoter:
                v = basket[s]
                close = market_data[s]['lastPrice']

                r = self.passorder(types, self.acc, s, xtconstant.FIX_PRICE, close - self.over_order_price, v,
                                   self._strategy_name, self.tzbz)
                if r > 0:
                    orders = type('order', (), {'order_id': r, 'stock_code': s, 'order_volume': v, 'order_status': 49,
                                                'order_type': xtconstant.STOCK_SELL, 'traded_volume': 0})
                    self.on_stock_order(orders)  # xtquant不推待报状态，这里自己造
                    self.ordered[r] = orders

                    del self.ready_to_sell[s]
                else:
                    pass
        else:
            for s in quoter:  # 买
                v = basket[s]
                close = market_data[s]['lastPrice']

                r = self.passorder(types, self.acc, s, xtconstant.FIX_PRICE, close + self.over_order_price, v,
                                   self._strategy_name, self.tzbz)
                if r > 0:
                    orders = type('order', (), {'order_id': r, 'stock_code': s, 'order_volume': v, 'order_status': 49,
                                                'order_type': xtconstant.STOCK_BUY, 'traded_volume': 0})
                    self.on_stock_order(orders)
                    self.ordered[r] = orders
                    del self.ready_to_buy[s]
                else:
                    pass

    def passorder(self, order_type, acct, stock_code, price_type, price, volume, strategy_name, tzbz):
        return self.t.order_stock(acct, stock_code, order_type, volume, price_type, price, strategy_name, tzbz)

    def on_stock_order(self, order):
        if not order.order_id in self.ordered:
            appLogger.error('order_id：{}不是配平委托，忽略'.format(order.order_id))
            return
        # print('委托回报',order.stock_code,order.order_volume,order.order_status)
        if order.order_status in [PART_CANCEL, CANCELED]:  # 部撤和已撤的，可放心再次下单
            if order.order_type == xtconstant.STOCK_BUY:
                self.ready_to_buy[order.stock_code] = order.order_volume - order.traded_volume
            elif order.order_type == xtconstant.STOCK_SELL:
                self.ready_to_sell[order.stock_code] = order.order_volume - order.traded_volume
            appLogger.debug('已撤,order_id:{},stock_code:{},order_volume:{},order_status:{}'.format(order.order_id,order.stock_code,order.order_volume,order.order_status))
            pass
        if order.stock_code not in self.ordered_status:
            self.ordered_status[order.stock_code] = [ORDER, order]
        else:
            if self.ordered_status[order.stock_code][0] != TRADE and (
            self.ordered_status[order.stock_code][1]).order_status < order.order_status:
                self.ordered_status[order.stock_code] = [ORDER, order]

    def on_stock_trade(self, trade):
        if not trade.order_id in self.ordered:
            appLogger.error('order_id：{}不是配平委托，忽略'.format(trade.order_id))
            return
        self.ordered_status[trade.stock_code] = [TRADE, trade]
        appLogger.debug('已成交：stock_code{},traded_volume{},direction：{}'.format(trade.stock_code, trade.traded_volume, trade.order_type))

    def on_cancel_error(self, cancel_error):

        if not cancel_error.order_id in self.ordered:
            appLogger.error('order_id：{}不是配平委托，忽略'.format(cancel_error.order_id))
            return
        appLogger.error('cancel error  error_msg:{},order_id:{}'.format(cancel_error.error_msg,cancel_error.order_id))
        order = self.t.query_stock_order(self.acc, cancel_error.order_id)
        self.wrong_status[order.stock_code] = [CANCEL_ERROR, order]
        pass

    def on_order_error(self, order_error):
        if not order_error.order_id in self.ordered:
            appLogger.error('order_id：{}不是配平委托，忽略'.format(order_error.order_id))
            return
        order = self.t.query_stock_order(self.acc, order_error.order_id)
        if not order:  # 这个似乎是不安全的代替方案，暂时query的结果是空，需要确定问题所在
            order = self.ordered[order_error.order_id]
        if order.order_type == xtconstant.STOCK_BUY:
            self.ready_to_buy[order.stock_code] = order.order_volume
        elif order.order_type == xtconstant.STOCK_SELL:
            self.ready_to_sell[order.stock_code] = order.order_volume
        appLogger.error('order_id:{}下单失败 error_msg:{}'.format(order_error.order_id, order_error.error_msg))


def parse_local_command(by_cmd=0):
    if by_cmd == 0:
        # cmd格式 空格分割，股票 数量，方向之间还用英文逗号隔开，并且他们之间一一对应
        # userdata路径 session_id account '000001.SZ,000002.SZ' '100,200' '1,0 ' 1：买  0：卖
        _ud_path, _ud_session, _account, _strategy_name, _stock, _volume, _direction = sys.argv[1:]
        _stock = _stock.split(',')
        _volume = [int(v) for v in _volume.split(',')]
        _direction = [int(v) for v in _direction.split(',')]
        # 检查参数完整性：
        if not len(_stock) == len(_volume) == len(_direction):
            raise ValueError('不正确的传参,入参长度：stock：{}下单量：{}方向：{}'.format(len(_stock), len(_volume), len(_direction)))
    else:  # 测试时使用，
        _ud_path = 'D:\\miniqmt4\\userdata_mini'
        _ud_session = 123456
        _account = '1000000367'
        _account = '1000000004'
        _stock = ['000001.SZ', '000002.SZ', '600000.SH']
        _stock = ['600000.SH']
        _volume = [100, 300, 1000]
        _volume = [1000]
        _direction = [0, 0, 0]  # 1卖，0买
        _direction = [0]  # 1卖，0买
        _strategy_name = ''
    return _ud_path, int(_ud_session), _account, _strategy_name, _stock, _volume, _direction


if __name__ == "__main__":

    over_order_price = 0.02  # 超价设置
    loop = 5  # 轮询间隔，默认10s,实际实盘可能要更长
    try:
        t = StrategyMniQmtT0(over_order_price, True, loop)
        t.init()
        t.t.stop()
    except:
        msg = traceback.format_exc()
        appLogger.fatal('遇到意外的情况，配平结束')
        appLogger.fatal(msg)
        t.t.stop()
        sys.exit(-1)

